#ifndef CCTERM_H_
#define CCTERM_H_

#include <string>
#include <list>

using namespace std;

#include "Term.h"
#include "Symbol.h"
#include "Signature.h"
#include "NoEqClassException.h"


class CCTerm;
typedef CCTerm* EqClass;

typedef list<CCTerm*> PredList;


template<class T> class Signature;


class CCTerm : public Term
{
	
public:
	CCTerm(Symbol const* symbol, CCTerm** args,CCTerm* merge);
	/**
	 * Constructs a term representation having the given symbol and the given arguments.
	 * 
	 * @param symbol the operation symbol of the term
	 * @param args the arguments of the operation. Of course the number of arguments for the given
	 * operation symbol must match it's arity.
	 */
	CCTerm(Symbol const* symbol, CCTerm** args);
	virtual ~CCTerm();
	
	
	EqClass getEqClass() ;
	
	/**
	 * Joins the equivalence class of this term with the one of the given term.
	 */
	void unionWith(CCTerm* t);
	
	/**
	 * Provides the signature of this term.
	 */
	
	Signature<EqClass>* getSignature() const;
	
	/**
	 * Provides the predecessor list iff this represents a equivalence class, otherwise
	 * NULL. The predecessor list contains all those terms having an argument that is
	 * contained in the equivalence class representet by this class.
	 */
	
	PredList* getPredList() const;
	
	/**
	 * Provides the index-th argument of this term.
	 */
	
	CCTerm* operator[](int index) const;
	
	bool isEqClass() const;
	
	int getAuxMember() const;
	
	void setAuxMember(int aux);
	
	#ifdef WEIGHTED_UNION
	int getWeight() const;
	#endif
	

	/**
	 * Merges the given equivalence class into this one. NOTE: This method
	 * is only supposed to be called if this object represents a equivalence 
	 * class.
	 */


	void unionEqClass(EqClass eq);
	
private:
	
	/**
	 * The predecessor list contains all those terms having an argument that is
	 * contained in the equivalence class representet by this class.
	 */
	
	PredList* predList;
	
	/**
	 * The the upper neighbour in the equivalence class of this term or this if this is
	 * the root;
	 */
	CCTerm* equivClass;
	
	int auxMember;
	
	#ifdef WEIGHTED_UNION
	int weight;
	#endif
};


// -------------- inline definitions ----------------------------
	
#ifdef WEIGHTED_UNION
inline int CCTerm::getWeight() const {
	return auxMember;
}
#endif

inline bool CCTerm::isEqClass() const{
	return equivClass == this; 
}

inline int CCTerm::getAuxMember() const {
	return auxMember;
}
	
inline void CCTerm::setAuxMember(int aux) {
	this->auxMember = aux;
}

inline CCTerm* CCTerm::operator[](int index) const {
	return (CCTerm*)args[index];
}

inline PredList* CCTerm::getPredList() const {
	return predList;
}

inline EqClass CCTerm::getEqClass() {
	if(equivClass != this) {
		equivClass = equivClass->getEqClass();
	}
	return equivClass;
}
	
inline void CCTerm::unionEqClass(EqClass eq) {
	#ifdef DEBUG
	if(predList == NULL) {
		throw NoEqClassException("This is not an EqClass!");
	}
	if(eq->predList == NULL) {
		throw NoEqClassException("Argument is not an EqClass!");
	}
	#endif
	#ifdef WEIGHTED_UNION
	weight += eq->weight;
	#endif
	eq->equivClass = this;
	predList->splice(predList->end(), *(eq->predList));
	delete eq->predList;
}
	
inline void CCTerm::unionWith(CCTerm* t) {
	getEqClass()->unionEqClass(t->getEqClass());
}
#endif /*CCTERM_H_*/
