#ifndef SIGNATURE_H_
#define SIGNATURE_H_

#include "Symbol.h"
#include "CCTerm.h"

#include <sstream>
#include <ext/hash_map>

using namespace std;

class CCTerm;
typedef CCTerm* EqClass;

/**
 * This class represents a equivalence class signature, witch is an abstraction from an term
 * representation as it contains the operation symbol but instead of the arguments it only
 * contains the respective equivalence classes.
 * 
 * @author Patrick Bahr
 */

template<class T> class Signature
{
public:
	//Signature(Symbol* symbol, int signature[]);
	/**
	 * Constructs a signature. NOTE: To intialize this signature completely the argument's
	 * equivalence classes must be set using the [] operator on this class.
	 * 
	 * @param symbol the symbol of this signature.
	 */
	Signature(Symbol const* symbol);
	virtual ~Signature();
	
	/**
	 * Provides a hash value for this signature.
	 */
	
	size_t hash() const;
	
	/**
	 * Provides a human readable string representation of this signature.
	 */
	 
	string toString() const;
	/**
	 * Provides the symbol of this signature
	 */
	Symbol const* getSymbol() const;
	
	/**
	 * Gets the array containig the equivalence classes of the arguments
	 */
	T const* getSignature() const;
	
	/**
	 * Provides the arity of this signature.
	 */
	int getArity() const;
	
	/**
	 * Enables direct acces to the signature.
	 * 
	 * @return the equvalence class at position index
	 */
	T& operator[](const int index) const;
	
	bool operator==(const Signature<T>& s) const;
	bool operator<(const Signature<T>& s) const;
	
private:

Symbol const*const symbol;
T* signature;
};

// ----------------------------- inline members ------------------------------------

template<class T> inline size_t Signature<T>::hash() const{
	size_t ret = symbol->hash();
	for(int i=0;i<symbol->getArity();i++) {
		ret ^= (size_t)signature[i];
	}
	return ret;
}

template<class T> inline int Signature<T>::getArity() const {
	return symbol->getArity();
}

template<class T> inline Symbol const* Signature<T>::getSymbol() const {
	return symbol;
}

template<class T> inline T const* Signature<T>::getSignature() const {
	return signature;
}

template<class T> inline T& Signature<T>::operator[](const int index) const{
	return signature[index];
}

template<class T> inline bool Signature<T>::operator<(const Signature<T>& s) const {
	if(*symbol < *(s.symbol)) {
		return true;
	} else if(*symbol == *(s.symbol)) {
		for(int i=0;i<symbol->getArity();i++) {
			if(signature[i] < s[i]) {
				return true;
			} else if(signature[i] != s[i]) {
				return false;
			}
		}
	}
	return false;
}

template<class T> inline bool Signature<T>::operator==(const Signature<T>& s) const {
	if(*symbol == *(s.symbol)) {
		for(int i=0;i<symbol->getArity();i++) {
			if(signature[i] != s[i]) {
				return false;
			}
		}
		return true;
	}
	return false;
}

// --------------------------- end of inline definitions --------------------------------------

// some definitions to make the hash_map implementation work for a key of type Signature* 

namespace std {
	template<> struct equal_to< Signature<EqClass>* > {
		bool operator()(const Signature<EqClass>* s ,const Signature<EqClass>* t ) const {
			if(s->getSymbol() == t->getSymbol()) {
				for(int i=0;i<s->getArity();i++) {
					if((*s)[i] != (*t)[i]) {
						return false;
					}
				}
				return true;
			} else {
				return false;
			}
		}
	};
	
	template<> struct equal_to< Signature<int>* > {
		bool operator()(const Signature<int>* s ,const Signature<int>* t ) const {
			if(s->getSymbol() == t->getSymbol()) {
				for(int i=0;i<s->getArity();i++) {
					if((*s)[i] != (*t)[i]) {
						return false;
					}
				}
				return true;
			} else {
				return false;
			}
		}
	};
};

namespace __gnu_cxx {
	template<> struct hash< Signature<EqClass>* > {
		size_t operator()(const Signature<EqClass>* sig ) const {
			return  sig->hash();
		}
	};
	
	template<> struct hash< Signature<int>* > {
		size_t operator()(const Signature<int>* sig ) const {
			return  sig->hash();
		}
	};
};

#endif /*SIGNATURE_H_*/
