#ifndef THEORY_H_
#define THEORY_H_

#include <iostream>
#include <ext/hash_map>
#include <list>
#include <set>

#include "Equation.h"
#include "Signature.h"
#include "CCTerm.h"
#include "parsing.h"
#include "TheoryParseException.h"
#include "SigTab.h" 

using namespace std;
using namespace __gnu_cxx;

// some typedefs for the data structures used here

typedef hash_map<Signature<EqClass>*,EqClass> SignatureTable;
typedef list<CCTerm*> Pending;
typedef list< pair<EqClass,EqClass> > Combine;
typedef Equation<CCTerm> CCEquation;

/**
 * 
 * 
 * @author Patrick Bahr
 */
class Theory
{
public:
	virtual ~Theory();
	
	/**
	 * Provides a human readable string representation of the signature table.
	 */
	
	string printSignatureTable() const;
	
	/**
	 * Provides a human readable string representaion of the predecessor list.
	 */
	string printPredList() const;
	
	SigTab* copySigTab() const;
	
	/**
	 * Provides the number of equations in this theory.
	 */
	 
	int getAxiomsCount() const;
	
	/**
	 * Provides the size of this theory.
	 */
	
	int size() const;
	
	/**
	 * Parses a the content of a given input stream containing equations to build
	 * a theorem.
	 * 
	 * @param input an input stream that contains the equations.
	 */
	
	static Theory* parse(istream& input);
	
	/**
	 * Builds the congruence closure of this theory.
	 */

	void buildClosure();
	
private:

	static void swap(EqClass& c1, EqClass& c2);

	/**
	 * Combines all pairs in the combine list using the combinePair() method, merging
	 * the equivalence class having the shorter predecessor list to the other equivalence
	 * class.
	 */

	void combineAll(Combine& combine,Pending& pending);
	
	/**
	 * Tries to insert the given term to the signature table. If there is already a
	 * term in the signature table having the same signature that term together
	 * in pair with the given term is added to the combine list.
	 */

	void insertToSignatureTable(CCTerm* t, Combine& combine);
	
	/**
	 * Traverses the given term initializing the signature table using the
	 * insertToSignatureTable() method and initializes the predecessor lists
	 * of the respective equivalence classes.
	 */

	void traverse(CCTerm* t,Combine& combine);
	
	/**
	 * Private constructor to prevent construction. Use the static member
	 * function Theory::parse() to get a theory.
	 */

	Theory();
	
	/**
	 * The ariyty table keeps track of the arity of the operation symbols, in order
	 * to guarantee that one symbol is used with one arity only.
	 */
	ArityTable arityTable;
	
	/**
	 * The size of the array axioms that contains the equation defining this
	 * theory.
	 */
	
	int axiomsCount;
	
	/**
	 * An array of equations that define this theory.
	 */
	
	CCEquation** axioms;
	
	SignatureTable signatureTable;
};

// -------------- inline definitions -----------------------------------

inline int Theory::getAxiomsCount() const {
	return axiomsCount;
}

inline int Theory::size() const {
	int s = 0;
	for(int i=0;i<axiomsCount;i++) {
		s+=axioms[i]->size();
	}
	return s;
}

inline void Theory::swap(EqClass& c1, EqClass& c2) {
	#ifdef WEIGHTED_UNION
	if(c1->getWeight() > c2->getWeight()) {
	#else
	if(c1->getPredList()->size() > c2->getPredList()->size()) {
	#endif
			EqClass& c = c1;
			c1 = c2;
			c2 = c;
	}
}


inline  void Theory::combineAll(Combine& combine,Pending& pending) {
	for(Combine::iterator it = combine.begin();it!=combine.end();it++) {
		EqClass c1 = (*it).first, c2 = (*it).second;
		swap(c1,c2);
		PredList* c1list = c1->getPredList();
		PredList::iterator c1listEnd = c1list->end();
		if (c1->isEqClass() && c2->isEqClass()) {
			for(PredList::iterator it = c1list->begin();it!=c1listEnd;it++) {
				Signature<EqClass>* sig = (*it)->getSignature();
				signatureTable.erase(sig);
				delete sig;
				pending.push_back((*it));
			}
			// union
			c2->unionEqClass(c1);
		}
	}
}

inline void Theory::insertToSignatureTable(CCTerm* t, Combine& combine) {
	Signature<EqClass>* sig = t->getSignature();
	SignatureTable::iterator it = signatureTable.find(sig);
	EqClass c = t->getEqClass();
	if(it == signatureTable.end()) {
		signatureTable.insert(make_pair(sig,c));
	} else {
		delete sig;
		if(c != it->second)
			combine.push_back(make_pair(c,it->second));
	}
}

#endif /*THEORY_H_*/
